%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>

#include "log.h"
#include "tools.h" /* provides str */
#include "filters.h"
#include "validate.h"
#include "destination.h" /* provides destination */
#include "conf-tools.h" /* needs str/destination, provides interface */

#define YYSTYPE char*
#define YYERROR_VERBOSE yes
#define CLEAR(X) if (X!=NULL) sfree(X); X=NULL;
#define REQUOTE(X) if (X!=NULL) sfree(X); X=stripquotes(yytext);
#define REASSIGN(X) if (X!=NULL) sfree(X); X=sstrdup(yytext);

    extern FILE *yyin;
    extern char *yytext;
    int yy_line, yylex();
    void yyerror(char *msg);

    int config_returncode;
    char *config_filename;
    config *current_config; /* initialised before yyparse */
    unsigned long config_timestamp=0;

    int config_port=-1;
    int config_maxwait=-1;
    int config_sendsize=-1;
    int config_log_level=-1;
    int config_data_rows=-1;
    int config_proxy_port=-1;
    int config_output_delay=-1;
    int config_output_maxlines=-1;
    int config_sensor_frequency=-1;
    int config_output_type=out_type_unknown;

    char *config_db_host=NULL;
    char *config_db_user=NULL;
    char *config_db_password=NULL;
    char *config_db_name=NULL;

    char *config_role_member_name=NULL;
    char *config_role_name=NULL;

    char *config_user_name=NULL;
    char *config_user_password=NULL;

    char *config_url=NULL;
    char *config_log_file=NULL;
    char *config_data_name=NULL;
    char *config_data_file=NULL;
    char *config_password=NULL;
    char *config_proxy_name=NULL;
    char *config_proxy_user=NULL;
    char *config_ip_address=NULL;
    char *config_output_name=NULL;
    char *config_output_role=NULL;
    char *config_sensor_name=NULL;
    char *config_output_query=NULL;
    char *config_output_heading=NULL;
    char *config_sensor_command=NULL;
    char *config_proxy_password=NULL;
    char *config_dependency_table=NULL;
    char *config_sensor_parameter=NULL;
    char *config_sensor_filter_name=NULL;
    char *config_machine_filter_name=NULL;
    char *config_dependency_signature=NULL;

    role *config_roles=NULL;
    role_member *config_role_members=NULL;

    user *config_users=NULL;
    user *config_identity=NULL;

    out_struct *config_outputs=NULL;
    sensor_filter *config_sensor_filters=NULL;
    machine_filter *config_machine_filters=NULL;
    out_dependency *config_output_dependencies=NULL;

    void config_cleanup()
	{
	    config_port=-1;
	    config_maxwait=-1;
	    config_sendsize=-1;
	    config_log_level=-1;
	    config_data_rows=-1;
	    config_proxy_port=-1;
	    config_output_delay=-1;
	    config_output_maxlines=-1;
	    config_sensor_frequency=-1;
	    config_output_type=out_type_unknown;

#define STRC(x) if (x!=NULL) { sfree(x); x=NULL; };
	    STRC(config_db_host);
	    STRC(config_db_user);
	    STRC(config_db_password);
	    STRC(config_db_name);

	    STRC(config_role_member_name);
	    STRC(config_role_name);

	    STRC(config_user_name);
	    STRC(config_user_password);

	    STRC(config_url);
	    STRC(config_log_file);
	    STRC(config_data_name);
	    STRC(config_data_file);
	    STRC(config_password);
	    STRC(config_proxy_name);
	    STRC(config_proxy_user);
	    STRC(config_ip_address);
	    STRC(config_output_name);
	    STRC(config_output_role);
	    STRC(config_sensor_name);
	    STRC(config_output_query);
	    STRC(config_output_heading);
	    STRC(config_sensor_command);
	    STRC(config_proxy_password);
	    STRC(config_dependency_table);
	    STRC(config_sensor_parameter);
	    STRC(config_sensor_filter_name);
	    STRC(config_machine_filter_name);
	    STRC(config_dependency_signature);

	    cleanup_roles(&config_roles);
	    cleanup_role_members(&config_role_members);
	    cleanup_users(&config_users);
	    cleanup_users(&config_identity);
	    cleanup_outputs(&config_outputs);
	    cleanup_sensor_filters(&config_sensor_filters);
	    cleanup_machine_filters(&config_machine_filters);
	    cleanup_output_dependencies(&config_output_dependencies);
	}

    void yyerror(char *msg)
	{
	    extern int yy_line;
	    static unsigned long yyerror_time=0;
	    log((1,&yyerror_time,"%s[%d]: %s.",config_filename,yy_line,msg));
	    config_returncode=false;
	}

    void check_length(char *buf, char *dsc, int factor)
	{
	    if (strlen(buf)>factor*VALIDATE_LABELLEN)
	    {
		static unsigned long yylength_time=0;
		log((1,&yylength_time,"config error: %s '%s' is longer than %d chars.",dsc,buf,factor*VALIDATE_LABELLEN));
		config_returncode=false;
	    };
	}

    config *fetch_config(const char *filename, config_type_enum tpe)
	{
	    struct stat sb;
	    config *res=NULL;

#ifdef YYDEBUG
	    extern int yydebug;
	    yydebug=1;
#endif

	    sb.st_mtime=0;
	    stat(filename,&sb);

	    if (sb.st_mtime>config_timestamp)
	    {
		FILE *src;

		config_returncode=true;
		config_timestamp=sb.st_mtime;

		config_filename=(char*)filename;
		log((7,NULL,"reading config file '%s'.",filename));

		config_cleanup();

		src = fopen(filename,"r");
		if (src==NULL)
		{
		    char cwdbuf[1000];
		    static unsigned long configfopen_time=0;
		    getcwd(cwdbuf,sizeof(cwdbuf));
		    log((1,&configfopen_time,"error reading file '%s' in '%s'.",filename,cwdbuf));
		    config_returncode=false;
		}
		else
		{
		    int yyparse();
		    void yyrestart(FILE*);

		    yy_line=1;
		    yyrestart(src);
		    res=new_config();
		    current_config=res;

		    yyparse();
		    fclose(src);

		    if (res->config_type!=tpe)
		    {
			char *tpe;
			static unsigned long configtype_time=0;
			config_returncode=false;
			switch(res->config_type)
			{
			    case ct_watcher: tpe="watcher"; break;
			    case ct_collector: tpe="collector"; break;
			    case ct_sender: tpe="sender"; break;
			    case ct_receiver: tpe="receiver"; break;
			    case ct_refinery: tpe="refinery"; break;
			    case ct_server: tpe="server"; break;
			    case ct_unknown: default: tpe="unknown"; break;
			};
			log((1,&configtype_time,"could not read config file (tpe=%s).",tpe));
		    };

		    if (config_returncode==true)
		    {
			log((7,NULL,"file '%s' has been parsed successfully.",filename));
		    }
		    else
		    {
			free_config(res);
			current_config=NULL;
			res=NULL;
		    };
		};
	    };
	    return res;
	}
%}

%token BLANK
%token EQUALS
%token IP_ADDRESS
%token LBRACE
%token COMMA
%token LFEED
%token NUMBER
%token RBRACE
%token SEMICOLON
%token COMMENT
%token STRING
%token SYMBOL

%token T_COLLECTOR_CONFIG
%token T_DIRECTORY_FREQUENCY
%token T_OFFER
%token T_MEMBERS
%token T_SERVER_CONFIG
%token T_COMMAND
%token T_DB_HOST
%token T_DB_NAME
%token T_DB_PASSWORD
%token T_DB_USER
%token T_DELAY
%token T_MAX_LINES
%token T_DEPENDENCIES
%token T_DEPENDENCY
%token T_DESTINATION
%token T_DESTINATIONS
%token T_FILTER
%token T_FREQUENCY
%token T_HEADING
%token T_HISTOGRAM
%token T_IDENTITY
%token T_INPUT_TIMEOUT
%token T_INTERFACE
%token T_LOG_FILE
%token T_LOG_LEVEL
%token T_MACHINE
%token T_MACHINES
%token T_MAXWAIT
%token T_NAME
%token T_ROLE
%token T_OUTPUT
%token T_OUTPUTS
%token T_PARAMETER
%token T_PASSWORD
%token T_PORT
%token T_PROXY
%token T_QUERY
%token T_RECEIVER_CONFIG
%token T_REFINERY_CONFIG
%token T_SCALAR
%token T_SENDER_CONFIG
%token T_SENDSIZE
%token T_SENSOR
%token T_SENSORS
%token T_SIGNATURE
%token T_SITE
%token T_SOURCE
%token T_TABLE
%token T_VIEWER_1_CONFIG
%token T_VIEWER_2_CONFIG
%token T_TYPE
%token T_URL
%token T_INPUT
%token T_VIEWER_INPUTS
%token T_DATA_ROWS
%token T_DATA_FREQ
%token T_DATA_FILE
%token T_DATA_NAME
%token T_USER
%token T_USERS
%token T_ROLES
%token T_WATCHER_CONFIG

%%

/* ---------------------------------------------------------------------- */

input: wss config_file

config_file:	watcher_config
	|	collector_config
	|	sender_config
	|	receiver_config
	|	refinery_config
	|	server_config
	|	viewer_1_config
	|	viewer_2_config
;

password_name: STRING { REQUOTE(config_password); }
;

password: T_PASSWORD wss EQUALS wss password_name wss SEMICOLON
;

site_name: SYMBOL
	{
	    check_length(yytext,"site name",1);
	    current_config->site_name=sstrdup(yytext);
	}
;

site: T_SITE wss EQUALS wss site_name wss SEMICOLON
;

machine_name: SYMBOL
	{
	    check_length(yytext,"machine name",1);
	    current_config->machine_name=sstrdup(yytext);
	}
;

machine: T_MACHINE wss EQUALS wss machine_name wss SEMICOLON
;

input_to_num: NUMBER
	{ current_config->input_timeout=atoi(yytext); }
;

input_timeout: T_INPUT_TIMEOUT wss EQUALS wss input_to_num wss SEMICOLON
;

ip_address: IP_ADDRESS
	{ REASSIGN(config_ip_address); }
;

interface: T_INTERFACE wss EQUALS wss ip_address wss SEMICOLON
;

port: T_PORT wss EQUALS wss port_number wss SEMICOLON
;

port_number: NUMBER
	{ config_port=atoi(yytext); }
;

url_text: STRING { REQUOTE(config_url); }
;

url: T_URL wss EQUALS wss url_text wss SEMICOLON
;

user_name_text: STRING { REQUOTE(config_user_name); }
;

user_password_text: STRING { REQUOTE(config_user_password); }
;

identity: T_IDENTITY wss EQUALS wss LBRACE wss
		T_NAME wss EQUALS wss user_name_text wss SEMICOLON wss
		T_PASSWORD wss EQUALS wss user_password_text wss SEMICOLON wss
RBRACE wss SEMICOLON
{
    if (config_identity!=NULL) cleanup_users(&config_identity);
    config_identity=new_user(config_user_name,config_user_password,NULL);
}
;

user: T_USER wss EQUALS wss LBRACE wss
		T_NAME wss EQUALS wss user_name_text wss SEMICOLON wss
		T_PASSWORD wss EQUALS wss user_password_text wss SEMICOLON wss
RBRACE wss SEMICOLON
{ config_users=new_user(config_user_name,config_user_password,config_users); }
;

user_list: user | user_list wss user;

users: T_USERS wss EQUALS wss LBRACE wss user_list wss RBRACE wss SEMICOLON
;

role_member: STRING
{
    REQUOTE(config_role_member_name);
    config_role_members=new_role_member(config_role_member_name,config_role_members);
}
;

role_member_list: role_member | role_member_list wss COMMA wss role_member;

role_name_text: STRING { REQUOTE(config_role_name); }
;

role: T_ROLE wss EQUALS wss LBRACE wss
		T_NAME wss EQUALS wss role_name_text wss SEMICOLON wss
		T_MEMBERS wss EQUALS wss LBRACE wss role_member_list wss RBRACE wss SEMICOLON wss
RBRACE wss SEMICOLON
{
    config_roles=new_role(config_role_name,config_role_members,config_roles);
    config_role_members=NULL;
}
;

role_list: role | role_list wss role;

roles: T_ROLES wss EQUALS wss LBRACE wss role_list wss RBRACE wss SEMICOLON;

/* ---------------------------------------------------------------------- */

watcher_config:
	T_WATCHER_CONFIG wss
	log_file wss
	log_level wss
	machine wss
	identity wss
	watcher_dst wss
	input_timeout wss
	watcher_sensors wss
	{
	  current_config->config_type=ct_watcher;
	  current_config->log_file=sstrdup(config_log_file);
	  current_config->log_level=config_log_level;
	  current_config->identity=config_identity; config_identity=NULL;
	}
;

watcher_dst: T_DESTINATION wss EQUALS wss
	LBRACE wss interface wss port wss RBRACE wss SEMICOLON
	{
		current_config->dst=new_interface(config_ip_address,config_port);
	}
;

watcher_sensors:
	T_SENSORS wss EQUALS wss LBRACE wss sensor wss sensors RBRACE SEMICOLON
;

sensors: /* empty */
	| sensor wss sensors
;

sensor: T_SENSOR wss EQUALS wss LBRACE wss
		sensor_name wss
		sensor_frequency wss
		sensor_command wss
		optional_sensor_parameter
		RBRACE SEMICOLON
	{
	    current_config->sensor_list=
		add_sensor(current_config->sensor_list,
			   config_sensor_frequency,
			   config_sensor_name,
			   config_sensor_command,
			   config_sensor_parameter);
	    CLEAR(config_sensor_parameter);
	}
;

sensor_name_symbol: SYMBOL
	{
	    check_length(yytext,"sensor name",1);
	    REASSIGN(config_sensor_name);
	}
;

sensor_name: T_NAME wss EQUALS wss sensor_name_symbol wss SEMICOLON
;

sensor_freq_number: NUMBER { config_sensor_frequency=atoi(yytext); }
;

sensor_frequency: T_FREQUENCY wss EQUALS wss sensor_freq_number wss SEMICOLON
;

sensor_command_string: STRING { REQUOTE(config_sensor_command); }
;

sensor_command: T_COMMAND wss EQUALS wss sensor_command_string wss SEMICOLON
;

sensor_parameter_string: STRING
	{ REQUOTE(config_sensor_parameter); }
;

sensor_parameter: T_PARAMETER wss EQUALS wss
	sensor_parameter_string wss SEMICOLON
;

optional_sensor_parameter: /* empty */ | sensor_parameter wss;

/* ---------------------------------------------------------------------- */

collector_config:
	T_COLLECTOR_CONFIG wss
	log_file wss
	log_level wss
	users wss
	source_port wss
	input_timeout wss
	{
	  current_config->config_type=ct_collector;
	  current_config->log_file=sstrdup(config_log_file);
	  current_config->log_level=config_log_level;
	  current_config->users=config_users; config_users=NULL;
	}
;

source_port: T_SOURCE wss EQUALS wss
	LBRACE wss interface wss port wss RBRACE wss SEMICOLON
	{ current_config->src=new_interface(config_ip_address,config_port); }
;

/* ---------------------------------------------------------------------- */

sender_config:
	T_SENDER_CONFIG wss
	log_file wss
	log_level wss
	site wss
	destinations wss
	{
	  current_config->config_type=ct_sender;
	  current_config->log_file=sstrdup(config_log_file);
	  current_config->log_level=config_log_level;
	}
;

destinations: T_DESTINATIONS wss EQUALS wss LBRACE wss destination_list wss RBRACE SEMICOLON
;

destination_list: sender_dst
	| destination_list wss sender_dst
;

proxy_name_string: STRING { REQUOTE(config_proxy_name); }
;

proxy_name: T_NAME wss EQUALS wss proxy_name_string wss SEMICOLON
;

proxy_port: T_PORT wss EQUALS wss proxy_port_number wss SEMICOLON
;

proxy_port_number: NUMBER
	{ config_proxy_port=atoi(yytext); }
;

proxy_user: STRING { REQUOTE(config_proxy_user); }
;

proxy_password: STRING { REQUOTE(config_proxy_password); }
;

optional_proxy_auth:
	/* empty */
	|
	T_USER wss EQUALS wss proxy_user wss SEMICOLON
	wss
	T_PASSWORD wss EQUALS wss proxy_password wss SEMICOLON
	wss
;

optional_proxy:
		/* empty */
	|
		T_PROXY wss
		EQUALS wss
		LBRACE wss
		proxy_name wss
		proxy_port wss
		optional_proxy_auth
		RBRACE wss
		SEMICOLON wss
;

sensor_filter: SYMBOL
	{
	    check_length(yytext,"sensor filter",1);
	    REASSIGN(config_sensor_filter_name);
	}
;

sensor_filter_struct:
	T_FILTER wss EQUALS wss LBRACE wss
	T_MACHINE wss EQUALS wss machine_filter wss SEMICOLON wss
	T_SENSOR wss EQUALS wss sensor_filter wss SEMICOLON wss
	RBRACE wss SEMICOLON
	{
		config_sensor_filters=
		new_sensor_filter(config_sensor_filters,
				  config_machine_filter_name,
				  config_sensor_filter_name);
	}
;

sensor_filter_list:
		sensor_filter_struct
	|
		sensor_filter_list wss sensor_filter_struct
;

optional_sensor_filter_list: /* empty */ | sensor_filter_list wss
;

sensor_filters: T_SENSORS wss EQUALS wss LBRACE wss optional_sensor_filter_list RBRACE wss SEMICOLON
;

machine_filter: SYMBOL
	{
	    check_length(yytext,"machine filter",1);
	    REASSIGN(config_machine_filter_name);
	}
;

machine_filter_struct:
	T_FILTER wss EQUALS wss LBRACE wss
	T_MACHINE wss EQUALS wss machine_filter wss SEMICOLON wss
	RBRACE wss SEMICOLON
	{
		config_machine_filters=
		new_machine_filter(config_machine_filters,
				   config_machine_filter_name);
	}
;

machine_filter_list: machine_filter_struct
	| machine_filter_list wss machine_filter_struct
;

optional_machine_filter_list: /* empty */ | machine_filter_list wss
;

machine_filters: T_MACHINES wss EQUALS wss LBRACE wss optional_machine_filter_list RBRACE wss SEMICOLON
;

maxwait_number: NUMBER
	{ config_maxwait=atoi(yytext); }
;

maxwait: T_MAXWAIT wss EQUALS wss maxwait_number wss SEMICOLON
;

sendsize_number: NUMBER
	{ config_sendsize=atoi(yytext); }
;

sendsize: T_SENDSIZE wss EQUALS wss sendsize_number wss SEMICOLON
;

sender_dst: T_DESTINATION wss EQUALS wss
	LBRACE wss
        maxwait wss
	sendsize wss
	password wss
	optional_proxy
	url wss
	sensor_filters wss
	machine_filters wss
	RBRACE wss
	SEMICOLON
	{
	    destination *x = new_destination(current_config->destination_list,
					     config_url,
					     config_sendsize,config_maxwait);
	    if (x==NULL)
	    {
		cleanup_sensor_filters(&config_sensor_filters);
		cleanup_machine_filters(&config_machine_filters);
	    }
	    else
	    {
		x->password=sstrdup(config_password);

		if (config_proxy_name!=NULL)
		{
		    x->proxy_name=sstrdup(config_proxy_name);
		    x->proxy_port=config_proxy_port;
		    if (config_proxy_user!=NULL)
		    {
			char credentials[1000];
			x->proxy_user=sstrdup(config_proxy_user);
			x->proxy_password=sstrdup(config_proxy_password);
			sprintf(credentials,"%s:%s",config_proxy_user,config_proxy_password);
			x->credentials=sstrdup(base64_encode(credentials,strlen(credentials)));
		    };
		};

		x->sensor_filters=config_sensor_filters;
		config_sensor_filters=NULL;

		x->machine_filters=config_machine_filters;
		config_machine_filters=NULL;

		current_config->destination_list = x;
	    };
	}
;

/* ---------------------------------------------------------------------- */

receiver_config:
	T_RECEIVER_CONFIG wss
	log_file wss
	log_level wss
	site wss
	db_host wss
	db_name wss
	db_user wss
	optional_db_password
	password wss
	{
	    current_config->config_type=ct_receiver;
	    current_config->log_file=sstrdup(config_log_file);
	    current_config->log_level=config_log_level;
	    current_config->receiver_password=sstrdup(config_password);
	    current_config->db_host=sstrdup(config_db_host);
	    current_config->db_name=sstrdup(config_db_name);
	    current_config->db_user=sstrdup(config_db_user);
	    if (config_db_password!=NULL)
		current_config->db_password=sstrdup(config_db_password);
	}
;

log_file_text: STRING { REQUOTE(config_log_file); }
;

log_file: T_LOG_FILE wss EQUALS wss log_file_text wss SEMICOLON
;

log_level: T_LOG_LEVEL wss EQUALS wss log_level_val wss SEMICOLON
;

log_level_val: NUMBER { config_log_level=atoi(yytext); }
;

db_host: T_DB_HOST wss EQUALS wss db_host_val wss SEMICOLON
;

db_host_val: STRING { REQUOTE(config_db_host); }
;

db_name: T_DB_NAME wss EQUALS wss db_name_val wss SEMICOLON
;

db_name_val: STRING { REQUOTE(config_db_name); }
;

db_user: T_DB_USER wss EQUALS wss db_user_val wss SEMICOLON
;

db_user_val: STRING { REQUOTE(config_db_user); }
;

optional_db_password: /* empty */ | db_password wss
;

db_password: T_DB_PASSWORD wss EQUALS wss db_password_val wss SEMICOLON
;

db_password_val: STRING { REQUOTE(config_db_password); }
;

/* ---------------------------------------------------------------------- */

output_dependency_table: STRING { REQUOTE(config_dependency_table); }

output_dependency_sig: STRING { REQUOTE(config_dependency_signature); }

output_dependency_struct:
	T_DEPENDENCY wss EQUALS wss LBRACE wss
	T_TABLE wss EQUALS wss output_dependency_table wss SEMICOLON wss
	T_SIGNATURE wss EQUALS wss output_dependency_sig wss SEMICOLON wss
	RBRACE wss SEMICOLON
	{ config_output_dependencies=
	      new_dependency(
		  config_output_dependencies,
		  config_dependency_table,config_dependency_signature); }
;

output_dependency_list: output_dependency_struct
	| output_dependency_list wss output_dependency_struct
;

/* ---------------------------------------------------------------------- */

output_name_string: SYMBOL
	{
	    check_length(yytext,"output_name",4);
	    REASSIGN(config_output_name);
	}
;

output_name: T_NAME wss EQUALS wss output_name_string wss SEMICOLON
;

output_role_string: SYMBOL
	{
	    check_length(yytext,"output_role",1);
	    REASSIGN(config_output_role);
	}
;

output_role: T_ROLE wss EQUALS wss output_role_string wss SEMICOLON
;

output_type_scalar: T_SCALAR
	{ config_output_type=out_type_scalar; }
;

output_type_histogram: T_HISTOGRAM
	{ config_output_type=out_type_histogram; }
;

output_type_table: T_TABLE
	{ config_output_type=out_type_table; }
;

output_type_enum: output_type_scalar
	| output_type_histogram
	| output_type_table
;

output_type: T_TYPE wss EQUALS wss output_type_enum wss SEMICOLON
;

output_delay_number: NUMBER
	{ config_output_delay=atoi(yytext); }
;

output_delay: T_DELAY wss EQUALS wss output_delay_number wss SEMICOLON
;

output_maxlines_number: NUMBER
	{ config_output_maxlines=atoi(yytext); }
;

output_maxlines: T_MAX_LINES wss EQUALS wss output_maxlines_number wss SEMICOLON
;

output_heading_string: STRING
	{  REQUOTE(config_output_heading); }
;

output_heading: T_HEADING wss EQUALS wss output_heading_string wss SEMICOLON
;

output_dependencies: T_DEPENDENCIES wss EQUALS wss LBRACE wss
	output_dependency_list wss RBRACE wss SEMICOLON
;

output_query_string: STRING
	{  REQUOTE(config_output_query); }
;

output_query: T_QUERY wss EQUALS wss output_query_string wss SEMICOLON
;

output_struct:
	T_OUTPUT wss EQUALS wss LBRACE wss
	output_name wss
	output_role wss
	output_type wss
	output_delay wss
	output_maxlines wss
	output_heading wss
	output_dependencies wss
	output_query wss
	RBRACE wss SEMICOLON
	{
	    out_struct *x=new_output(config_outputs,
				     config_output_name,
				     config_output_role,
				     config_output_heading,
				     config_output_type,
				     config_output_delay,
				     config_output_maxlines,
				     config_output_query);
	    if (x==NULL)
	    {
		cleanup_output_dependencies(&config_output_dependencies);
	    }
	    else
	    {
		x->dependencies=config_output_dependencies;
		config_output_dependencies=NULL;
		config_outputs=x;
	    };
	}
;

output_list: output_struct
	| output_list wss output_struct
;

outputs: T_OUTPUTS wss EQUALS wss LBRACE wss output_list wss RBRACE wss SEMICOLON
;

/* ---------------------------------------------------------------------- */

refinery_config:
	T_REFINERY_CONFIG wss
	log_file wss
	log_level wss
	user wss
	db_host wss
	db_name wss
	db_user wss
	optional_db_password
	source_port wss
	input_timeout wss
	outputs wss
	{
	  current_config->config_type=ct_refinery;
	  current_config->log_file=sstrdup(config_log_file);
	  current_config->log_level=config_log_level;
	  current_config->users=config_users; config_users=NULL;
	  current_config->outputs=config_outputs; config_outputs=NULL;
	  current_config->db_host=sstrdup(config_db_host);
	  current_config->db_name=sstrdup(config_db_name);
	  current_config->db_user=sstrdup(config_db_user);
	  if (config_db_password!=NULL)
	      current_config->db_password=sstrdup(config_db_password);
	}
;

/* ---------------------------------------------------------------------- */

server_config:
	T_SERVER_CONFIG wss
	log_file wss
	log_level wss
	identity wss
	source_port wss
	input_timeout wss
	server_offer wss
	directory_frequency wss
	users wss
	roles wss
	{
	  current_config->config_type=ct_server;
	  current_config->log_file=sstrdup(config_log_file);
	  current_config->log_level=config_log_level;
	  current_config->identity=config_identity; config_identity=NULL;
	  current_config->users=config_users; config_users=NULL;
	  current_config->roles=config_roles; config_roles=NULL;
	}
;

server_offer: T_OFFER wss EQUALS wss
	LBRACE wss interface wss port wss RBRACE wss SEMICOLON
	{ current_config->dst=new_interface(config_ip_address,config_port); }
;

dir_freq_num: NUMBER { current_config->directory_frequency=atoi(yytext); };

directory_frequency:
	T_DIRECTORY_FREQUENCY wss EQUALS wss dir_freq_num wss SEMICOLON;

/* ---------------------------------------------------------------------- */

viewer_1_config:
	T_VIEWER_1_CONFIG wss
	log_file wss
	log_level wss
	identity wss
	source_port wss
	input_timeout wss
	data_freq wss
	viewer_inputs wss
	{
	  current_config->config_type=ct_viewer_1;
	  current_config->log_file=sstrdup(config_log_file);
	  current_config->log_level=config_log_level;
	  current_config->identity=config_identity; config_identity=NULL;
	}
;

data_freq_string: NUMBER { current_config->data_freq=atoi(yytext); };

data_freq: T_DATA_FREQ wss EQUALS wss data_freq_string wss SEMICOLON;


viewer_inputs: T_VIEWER_INPUTS wss EQUALS wss LBRACE wss viewer_input_list RBRACE wss SEMICOLON;

viewer_input_list: /* empty */ | viewer_input wss viewer_input_list;

viewer_input: T_INPUT wss EQUALS wss LBRACE wss data_name wss data_rows wss data_file wss RBRACE wss SEMICOLON
{
    current_config->viewer_inputs=new_viewer_input(config_data_name,config_data_rows,config_data_file,
						   current_config->viewer_inputs);
};

data_name_string: SYMBOL
	{
	    check_length(yytext,"data_name",4);
	    REASSIGN(config_data_name);
	};

data_name: T_DATA_NAME wss EQUALS wss data_name_string wss SEMICOLON;


data_rows_string: NUMBER { config_data_rows=atoi(yytext); };

data_rows: T_DATA_ROWS wss EQUALS wss data_rows_string wss SEMICOLON;


data_file_string: STRING { REQUOTE(config_data_file); };

data_file: T_DATA_FILE wss EQUALS wss data_file_string wss SEMICOLON;

/* ---------------------------------------------------------------------- */

viewer_2_config:
	T_VIEWER_2_CONFIG wss
	log_file wss
	log_level wss
	identity wss
	source_port wss
	input_timeout wss
	{
	  current_config->config_type=ct_viewer_2;
	  current_config->log_file=sstrdup(config_log_file);
	  current_config->log_level=config_log_level;
	  current_config->identity=config_identity; config_identity=NULL;
	}
;

/* ---------------------------------------------------------------------- */

ws:	BLANK
	| COMMENT
	| LFEED		{ yy_line++; }
;

wss:	/* empty */
	| ws wss
;

/* ---------------------------------------------------------------------- */

