/* -*- Mode:C++ -*- */

/* ---------------------------------------------------------------------- */

#include <time.h>
#include <ctype.h>
#include <stdio.h>
#include <getopt.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>

#include <netinet/in.h> /* for ntohl */

#include "log.h"
#include "tools.h"
#include "destination.h" /* provides destination */
#include "conf-tools.h" /* needs destination */

/* ---------------------------------------------------------------------- */

int main_fd=-1;

/* ---------------------------------------------------------------------- */

int md_cnt=0, md_avl=0;
typedef struct
{
    void *b;
    char *fn;
    int ln;
    int sz;
} mdebug;
mdebug *md_entries=NULL;

void md_add(void *r, int sz, const char *fn, int ln)
{
    mdebug x;
    if (md_cnt>=md_avl)
    {
	md_avl=(md_avl==0)?(128):(md_avl*2);
	md_entries=(mdebug*)realloc(md_entries,md_avl*sizeof(mdebug));
    };
    x.b=r; x.sz=sz; x.fn=(char*)fn; x.ln=ln; md_entries[md_cnt++]=x;
}

void md_del(void *r, const char *fn, int ln)
{
    int i;
    for (i=0; i<md_cnt; i++)
    {
	if (md_entries[i].b==r)
	{
	    md_entries[i] = md_entries[--md_cnt];
	    i=md_cnt+1;
	};
    };
    if (i<md_cnt+1)
    {
	fprintf(stderr,"md_del(%s,%d) without alloc\n",fn,ln);
	fflush(stderr);
    };
}

#ifdef MEMDEBUG
void md_lst(const char *fn, int ln)
{
    int i;

    log_nocleanup((7,NULL,"md_lst()"));
    for (i=0; i<md_cnt; i++)
	log_nocleanup((7,NULL,"%s[%d]=(%08lx,%d)",
		       md_entries[i].fn,md_entries[i].ln,
		       (unsigned long)md_entries[i].b,md_entries[i].sz));
}
#endif

void *int_smalloc(unsigned long s, const char *fn, int ln)
{
    void *r=malloc(s);
#ifdef MEMDEBUG
    md_add(r,s,fn,ln);
#endif

    if (r==NULL)
    {
	log((0,NULL,"%s[%d]: out of memory during malloc(%lu).",fn,ln,s));
	exit(1);
    };

    return r;
}

void *int_srealloc(void *b, unsigned long s, const char *fn, int ln)
{
    void *r;
    if (b==NULL)
    {
	r=int_smalloc(s,fn,ln);
    }
    else
    {
#ifdef MEMDEBUG
	md_del(b,fn,ln);
#endif
	r=realloc(b,s);
#ifdef MEMDEBUG
	md_add(r,s,fn,ln);
#endif
    };

    if (r==NULL)
    {
	log((0,NULL,"%s[%d]: out of memory during remalloc(%lu).",fn,ln,s));
	exit(1);
    };

    return r;
}

void int_sfree(void *b, const char *fn, int ln)
{
#ifdef MEMDEBUG
    md_del(b,fn,ln);
#endif
    if (b!=NULL) free(b);
}

char *int_sstrdup(const char *x, const char *f, int l)
{
    unsigned long sl=strlen(x)+1;
    char *res=int_smalloc(sl,f,l);
    strcpy(res,x);
    return res;
}

/* ---------------------------------------------------------------------- */

char *int_stripquotes(const char *x, const char *fn, int ln)
{
    int l;
    char *res;
    if (*x=='"') x++;
    res=int_sstrdup(x,fn,ln);
    l=strlen(res);
    if ( (l>0) && (res[l-1]=='"') ) res[l-1]=0;
    return res;
}

/* ---------------------------------------------------------------------- */

char *gmepoch_to_iso(unsigned long tme)
{
    static char res[30];
    static unsigned long gmtime_time=0;
    struct tm *ltime=gmtime((long*)&tme);

    tzset();

    if (ltime==NULL)
    {
	log((1,&gmtime_time,"couldn't get gmtime (res=NULL)."));
	res[0]=0;
    }
    else
    {
	sprintf(res,"%04d-%02d-%02d %02d:%02d:%02d-00",
		ltime->tm_year+1900,
		ltime->tm_mon+1,
		ltime->tm_mday,
		ltime->tm_hour,
		ltime->tm_min,
		ltime->tm_sec
	    );
    };

    return res;
}

char *iso_timestamp()
{
    return gmepoch_to_iso(time(NULL));
}

unsigned long iso_to_epoch(char *x)
{
    struct tm t;
    int year=0, month=0, day=0, hour=0, minute=0, second=0, offset=0;

    if (sscanf(x,"%4d-%2d-%2d %2d:%2d:%2d%3d",
	       &year,&month,&day,&hour,&minute,&second,&offset)==7)
    {
	t.tm_year=year-1900;
	t.tm_mon=month-1;
	t.tm_mday=day;
	t.tm_hour=hour+offset;
	t.tm_min=minute;
	t.tm_sec=second;
    }
    else
    {
	t.tm_year=70;
	t.tm_mon=0;
	t.tm_mday=1;
	t.tm_hour=0;
	t.tm_min=0;
	t.tm_sec=0;
    };

    return mktime(&t);
}

/* ---------------------------------------------------------------------- */

str *int_new_str(const char *fn, int ln)
{
    str *res=(str *)int_smalloc(sizeof(str),fn,ln);
    res->cnt=0;
    res->avl=0;
    res->buf=NULL;
    int_reset_str(res,fn,ln);
    return res;
}

void int_free_str(str *x, const char *fn, int ln)
{
    x->cnt=0;
    x->avl=0;
    if (x->buf!=NULL) sfree(x->buf);
    x->buf=NULL;
    if (x!=NULL) sfree(x);
}

void int_reset_str(str *x, const char *fn, int ln)
{
    if (x->avl==0)
    {
	x->avl=128;
	x->buf=(char*)int_smalloc(x->avl,fn,ln);
    };

    x->cnt=0;
    *(x->buf)=0;
}

int clen(str *x)
{
    return x->cnt;
}

char *cstr(str *x)
{
    (x->buf)[clen(x)]=0;
    return x->buf;
}

void int_add_str(str *x, char *add, const char *fn, int ln)
{
    int l=strlen(add);
    if (clen(x)+l >= (x->avl))
    {
	x->avl=clen(x)+l+1;
	x->buf=(char*)int_srealloc(x->buf,x->avl,fn,ln);
    };
    memcpy((x->buf)+clen(x),add,l+1);
    x->cnt = clen(x)+l;
}

void int_cpy_str(str *x, char *add, int count, const char *fn, int ln)
{
    int c=clen(x), l=count;
    if (c+l+1 >= (x->avl))
    {
	x->avl=c+l+1;
	x->buf=(char*)int_srealloc(x->buf,x->avl,fn,ln);
    };
    memcpy((x->buf)+c,add,l);
    (x->buf)[c+l]=0;

    x->cnt = c+l;
}

/* ---------------------------------------------------------------------- */

#define NOT_FOUND ((unsigned long)(-1))

unsigned long rev_search(char *buf, unsigned long cnt, char x)
{
    unsigned long res=NOT_FOUND;

    while (cnt>0)
    {
	if (buf[cnt]==x)
	{
	    res=cnt;
	    cnt=0;
	}
	else
	{
	    cnt--;
	};
    };

    return res;
}

static str *extract_line_s=NULL;
#define s extract_line_s
char *extract_first_line(char *buffer, int *bcnt)
{
    char *cr, *res=NULL;

    if (s==NULL) { s=new_str(); } else { reset_str(s); };

    cr=(buffer==NULL)?(NULL):(index(buffer,'\n'));
    if (cr==NULL)
    {
	res=NULL;
    }
    else
    {
	int l=cr-buffer+1;
	cpy_str(s,buffer,l); *bcnt-=l;
	memcpy(buffer,cr+1,*bcnt);
	buffer[*bcnt]=0;
	res=cstr(s);
    };

    return res;
}
char *extract_last_line(char *buffer, int *bcnt)
{
    char *res=NULL;
    unsigned long fst, sec;
    if (s==NULL) { s=new_str(); } else { reset_str(s); };

    sec=rev_search(buffer,*bcnt,'\n');
    if (sec++!=NOT_FOUND)
    {
	fst=(sec>1)?(rev_search(buffer,sec-2,'\n')):0;
	if (fst++!=NOT_FOUND)
	{
	    int len=sec-fst, rst=(*bcnt)-sec;
	    if (len>0) cpy_str(s,buffer+fst,len);
	    if (rst>0) memcpy(buffer+fst,buffer+sec,rst);
	    *bcnt-=len;
	    buffer[*bcnt]=0;
	    res=cstr(s);
	};
    };

    return res;
}
#undef s

/* ---------------------------------------------------------------------- */

volatile int end=false;

void my_handler(int s)
{
    switch(s)
    {
    case SIGINT:	end=true; log((2,NULL,"received INT signal.")); break;
    case SIGTERM:	end=true; log((2,NULL,"received TERM signal.")); break;
    default: break;
    };
}

void install_handler()
{
    signal(SIGINT,my_handler);
    signal(SIGTERM,my_handler);
    signal(SIGPIPE,SIG_IGN);
    signal(SIGCHLD,SIG_IGN);
}

/* ---------------------------------------------------------------------- */

void parse_commands(int argc, char **argv, char **cnf, char **log)
{
    char *optstring="c:dhl:";
    int o=getopt(argc,argv,optstring);

    while (o!=-1)
    {
	switch(o)
	{
	case 'c':
	    *cnf=optarg;
	    break;
	case 'h':
	    fprintf(stderr,"This is a sysmon component (C) 2001 by Oliver Seidel\n"); fflush(stderr);
	    fprintf(stderr,"usage: %s [-c <config-file>] [-h] [-l <log-file>]\n",argv[0]); fflush(stderr);
	    fprintf(stderr,"\n"); fflush(stderr);
	    fprintf(stderr,"\t-c permits the specification of a config file other than '%s.conf'.\n",argv[0]); fflush(stderr);
	    fprintf(stderr,"\t-h prints this help message.\n"); fflush(stderr);
	    fprintf(stderr,"\t-l permits the specification of a log file other than '%s.log'.\n",argv[0]); fflush(stderr);
	    fprintf(stderr,"\n"); fflush(stderr);
	    exit(1);
	case 'l':
	    *log=optarg;
	    break;
	};
	o=getopt(argc,argv,optstring);
    };
}

/* ---------------------------------------------------------------------- */

char *expect(char *buf, char *cmp)
{
    char *res=NULL;

    if ( (buf!=NULL) && (cmp!=NULL) )
    {
	int l=strlen(cmp);
	if (memcmp(buf,cmp,l)==0)
	    res=buf+l;
    };
    return res;
}

/* ---------------------------------------------------------------------- */

char *extract_int(char *buf, int *x)
{
    int num=0, fac=1;

    if (*buf=='-') { fac=-1; buf++; };

    while (isdigit(*buf))
    {
	num=10*num+(*buf)-'0';
	buf++;
    };

    *x = fac*num;

    return buf;
}

/* ---------------------------------------------------------------------- */

int split_url(char *url, char **hostname, int *port, char **path)
{
    int res=false;
    char *tmp=url;

    *hostname=NULL;
    *port=80;
    *path=NULL;

    tmp=expect(tmp,"http://");
    if (tmp!=NULL)
    {
	int end=false;
	char *host=tmp;
	while (end==false)
	{
	    if (isalnum(*tmp))
	    {
		tmp++;
	    }
	    else switch(*tmp)
	    {
	    case '.':
	    case '-':
	    case '_':
		tmp++;
		break;
	    default:
		end=true;
	    };
	};
	if (*tmp==':')
	{
	    *tmp=0;
	    tmp++;
	    tmp=extract_int(tmp,port);
	    *path=sstrdup(tmp);
	    *hostname=sstrdup(host);
	    res=true;
	}
	else if (*tmp=='/')
	{
	    char x=*tmp;
	    *tmp=0; *hostname=sstrdup(host);
	    *tmp=x; *path=sstrdup(tmp);
	    res=true;
	}
	else
	{
	    log((0,NULL,"error parsing url '%s'.\n",url));
	};
    };

    return res;
}

/* ---------------------------------------------------------------------- */

int base64_is_initialised=0;
char base64_reverse[256], base64_forward[] =
"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

void base64_initialise()
{
    int i, symbols=strlen(base64_forward);
    for (i=0; i<256; i++) base64_reverse[i] = -1;
    for (i=0; i<symbols; i++) base64_reverse[(int)base64_forward[i]] = i;
    base64_reverse['='] = 0;
    base64_is_initialised = 1;
}

static str *base64_encode_s=NULL;
#define s base64_encode_s
char *base64_encode(unsigned char *data, int count)
{
    int off, triple;
    unsigned long accumulator;

    if (s==NULL) s=new_str();

    reset_str(s);

    if (base64_is_initialised==0) base64_initialise();

    for (off=0, accumulator=0, triple=0; off<count; off++)
    {
	accumulator = (accumulator<<8) | data[off]; triple++;
	if (triple>=3)
	{
	    char buf[5];
	    buf[0]=base64_forward[(accumulator>>18)&0x3f];
	    buf[1]=base64_forward[(accumulator>>12)&0x3f];
	    buf[2]=base64_forward[(accumulator>> 6)&0x3f];
	    buf[3]=base64_forward[(accumulator>> 0)&0x3f];
	    buf[4]=0;
	    add_str(s,buf);
	    triple=0;
	    accumulator=0;
	};
    };

    if (triple!=0)
    {
	char buf[5];
	accumulator <<= (24 - (8*triple));
	buf[0]=base64_forward[(accumulator>>18)&0x3f];
	buf[1]=base64_forward[(accumulator>>12)&0x3f];
	if (triple==1)
	{
	    buf[2]='=';
	    buf[3]='=';
	}
	else
	{
	    buf[2]=base64_forward[(accumulator>> 6)&0x3f];
	    buf[3]='=';
	};
	buf[4]=0;
	add_str(s,buf);
    };

    return (cstr(s));
}

char *int_base64_decode(char *data,int *count, const char *fn, int ln)
{
    char c;
    static int avl=0;
    static char *res=NULL;

    unsigned long accumulator=0;
    int k=0, cnt=0, triple=0, prev;

    if (base64_is_initialised==0) base64_initialise();

    while (*data!=0)
    {
	prev=k; k = (int) *data;
	c=base64_reverse[k];
	if (c >= 0)
	{
	    accumulator = (accumulator<<6) + c; triple++;
	    if (triple>=4)
	    {
		if (avl<=cnt+4)
		{
		    avl=(2*avl)|128;
		    res=(char*)int_srealloc(res,avl,fn,ln);
		};
		res[cnt++]=(accumulator>>16)&0xff;
		res[cnt++]=(accumulator>> 8)&0xff;
		res[cnt++]=(accumulator>> 0)&0xff;
		triple=0;
		accumulator=0;
	    };
	};
	data++;
    };
    if (k   =='=') cnt--;
    if (prev=='=') cnt--;
    res[cnt]=0;
    *count=cnt;

    return res;
}

/* ---------------------------------------------------------------------- */

void tea_encipher(const unsigned long *v, unsigned long *w, const unsigned long *k)
{
    register unsigned long y=v[0],z=v[1],sum=0,delta=0x9e3779b9,n=32;
    while (n-->0)
    {
	y += (z << 4 ^ z >> 5) + (z ^ sum) + k[sum & 3];
	sum += delta;
	z += (y << 4 ^ y >> 5) + (y ^ sum) + k[sum >> 11 & 3];
    };
    w[0]=y; w[1]=z;
}

void cbc_encrypt(char *pwd, char *buf, unsigned long len)
{
    unsigned long ivec[2], data[2], key[4], off=0;

    memset(ivec,0,8);
    memset(key,0,16); strncpy((char*)key,pwd,16);

    while (off<len)
    {
	memcpy(data,buf+off,8);
	data[0]^=ivec[0];
	data[1]^=ivec[1];
	tea_encipher(data,data,key);
	memcpy(ivec,data,8);
	memcpy(buf+off,data,8);
	off+=8;
    };
}

char *mac_encrypt(char *buf, unsigned long len)
{
    int i;
    static char res[17];
    unsigned long ivec[2], data[2], key[4], off=0;

    memset(ivec,0,8);
    memset(key,0,16);

    while (off<len)
    {
	unsigned long rest=len-off;
	if (rest>7)
	{
       	    memcpy(data,buf+off,8);
	}
	else
	{
	    memcpy(data,buf+off,rest);
	    memset(((unsigned char *)(data))+rest,0,8-rest);
	};

	data[0]^=ivec[0];
	data[1]^=ivec[1];
	tea_encipher(data,data,key);
	memcpy(ivec,data,8);
	off+=8;
    };

    for (i=0; i<8; i++) sprintf(res+2*i,"%02x",((unsigned char *)(ivec))[i]);
    res[16]=0;

    return res;
}

void tea_decipher(const unsigned long *v, unsigned long *w, const unsigned long *k)
{
    register unsigned long y=v[0],z=v[1],sum=0xc6ef3720,delta=0x9e3779b9,n=32;
    while (n-->0)
    {
	z -= (y << 4 ^ y >> 5) + (y ^ sum) + k[sum >> 11 & 3];
	sum -= delta;
	y -= (z << 4 ^ z >> 5) + (z ^ sum) + k[sum & 3];
    };
    w[0]=y; w[1]=z;
}

void cbc_decrypt(char *pwd, char *buf, unsigned long len)
{
    unsigned long tmp[2], ivec[2], data[2], key[4], off=0;

    memset(ivec,0,8);
    memset(key,0,16); strncpy((char*)key,pwd,16);

    while (off<len)
    {
	memcpy(data,buf+off,8);
	memcpy(tmp,data,8);
	tea_decipher(data,data,key);
	data[0]^=ivec[0];
	data[1]^=ivec[1];
	memcpy(ivec,tmp,8);
	memcpy(buf+off,data,8);
	off+=8;
    };
}

/* ---------------------------------------------------------------------- */

void tools_cleanup()
{
    if (extract_line_s!=NULL)
    {
	free_str(extract_line_s);
	extract_line_s=NULL;
    };
    if (base64_encode_s!=NULL)
    {
	free_str(base64_encode_s);
	base64_encode_s=NULL;
    };
}

/* ---------------------------------------------------------------------- */

int is_ws(int ch)
{
    int res=false;
    switch(ch)
    {
    case ' ':
    case '\t':
    case '\n':
    case '\r':
	res=true;
	break;
    default:
	break;
    };
    return res;
}

/* ---------------------------------------------------------------------- */

int skip_ws(FILE *f)
{
    int c=fgetc(f);
    while ( (is_ws(c)==true) && (feof(f)==0) ) c=fgetc(f);
    return c;
}

/* ---------------------------------------------------------------------- */

char *ulong_to_dotted(unsigned long x)
{
    static char res[100];
    unsigned long y=ntohl(x);
    sprintf(res,"%lu.%lu.%lu.%lu",
	    (y>>24)&0xff,
	    (y>>16)&0xff,
	    (y>> 8)&0xff,
	    (y>> 0)&0xff
	    );
    return res;
}

/* ---------------------------------------------------------------------- */

char *tohex(char *buf, int max)
{
    char *res=0;
    int i, len=0;
    unsigned char *tmp=(unsigned char*)buf;

    if (len<3*max+20)
    {
	len=3*max+20;
	res=srealloc(res,len);
    };

    sprintf(res,"len=%8d ",max);
    for (i=0;i<max;i++) sprintf(res+13+3*i,"%02x ",tmp[i]);

    return res;
}

/* ---------------------------------------------------------------------- */

int write_string(str *s, int socket)
{
    int res=true;
    unsigned long written=0, to_write=clen(s);

    while ( (written<to_write) && (res==true) )
    {
	int rc=write(socket,cstr(s)+written,to_write-written);
	if (rc>-1)
	{
	    written+=rc;
	}
	else
	{
	    perror("write_string(s,socket)");
	    res=false;
	};
    };

    return res;
}

/* ---------------------------------------------------------------------- */

